/*********************************************************************/
/* announce.c                                                        */
/* function neblo_announce()                                         */
/* see COPYING in parent directory                                   */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


/*
 * the peer is announced (by a friend) that a "friendship" request
 * is about to arrive from a remote peer.
 * (command "announce").
 * insert a descriptor in the list of expected requests, if room available.
 * the announced "friendship" may be flagged as "new predecessor" if expected
 * to arrive from a new predecessor in the ring.
 */
int
neblo_announce
( reqst_t *reqst )
{

  unsigned long         saddr;
  struct in_addr        foreign_addr;
  unsigned int          foreign_port;
  unsigned int          predecessor;
  char                 *foreign_keymod;
  char                 *foreign_keyexp;
  RSA                  *foreign_pubkey;
  int                   rc = 0;


  _do_(_malloc_(foreign_keymod,KEYBUFLEN+SLACK),
       !foreign_keymod, "neblo_announce(): foreign_keymod malloc");
  _do_(_malloc_(foreign_keyexp,BUFLEN+SLACK),
       !foreign_keyexp, "neblo_announce(): foreign_keyexp malloc");

  sscanf(reqst->buf,"%*s %lu %u %u %50000s %100s",
         &saddr,&foreign_port,&predecessor,foreign_keymod,foreign_keyexp);

  /*
   * SECURITY: friendship announces may only come from friends
   */
  if (!is_friend(reqst->csock)) {

#ifdef DEBUG_F
    S(fprintf(stderr,"[%lu]\tneblo_announce(): announcer is not a friend\n",
              reqst->th_id));
#endif

    _free_(foreign_keymod); _free_(foreign_keyexp);
    _close_(&reqst->csock);
    return -1;
  }

  /*
   * SECURITY II: annouces flagged as "new predecessor" may only come from
   * current predecessor
   */
  if (predecessor && !from_predecessor(reqst->csock)) {

#ifdef DEBUG_F
    S(fprintf(stderr,
              "[%lu]\tneblo_announce(): ``new predecessor'' announce not coming from our current predecessor\n",
              reqst->th_id));
#endif

    _free_(foreign_keymod); _free_(foreign_keyexp);
    _close_(&reqst->csock);
    return -1;
  }

  foreign_addr.s_addr = saddr;

  _do_(_RSA_new_(foreign_pubkey),!foreign_pubkey,
       "neblo_announce(): RSA_new of foreign_pubkey");
  _convert_pubkey_(foreign_pubkey,foreign_keymod,foreign_keyexp,rc = -1);
  if (rc < 0) {

#ifdef DEBUG
    S(fprintf(stderr,"[%lu]\tneblo_announce(): bad foreign pubkey\n",
              reqst->th_id));
#endif

    _free_(foreign_keymod); _free_(foreign_keyexp);
    _RSA_free_(foreign_pubkey);
    _close_(&reqst->csock);
    return -1;
  }

  rc = insert_expd_reqst(foreign_addr,foreign_port,predecessor,foreign_pubkey,
                         reqst->th_id);

  _free_(foreign_keymod); _free_(foreign_keyexp);
  return 0;

}
