/*********************************************************************/
/* friendship.c                                                      */
/* function neblo_friendship()                                       */
/* see COPYING in parent directory                                   */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


/*
 *  This peer is asked to accept a permanent connection by the requestor.
 *  Command "friendship <ip_port> <kind>"
 *  where <kind> can be:
 *  0 (generic friend), 1 (friend is predecessor), 2 (friend is successor)
 */
int
neblo_friendship
( reqst_t *reqst )
{

  peer_t                *requestor;
  unsigned int           rport;
  unsigned int           pred_succ;
  peer_t                *f;
  int                    rc = 0;


  _do_(_calloc_(requestor,sizeof(peer_t)),
       !requestor, "neblo_friendship(): requestor calloc");

  sscanf(reqst->buf,"%*s %u %u",&rport,&pred_succ);
  requestor->port = (u_int16_t)rport;
  requestor->addr.s_addr = reqst->peer_addr.s_addr;

  /*  Ignore if already friend.  */

  if (is_friend(reqst->csock)) {

#ifdef DEBUG_F
    S(fprintf(stderr,
              "[%lu]\tfriendship req (peer %u.%u.%u.%u:%u): already friend\n",
              reqst->th_id,
              ((unsigned char *)&requestor->addr.s_addr)[0],
              ((unsigned char *)&requestor->addr.s_addr)[1],
              ((unsigned char *)&requestor->addr.s_addr)[2],
              ((unsigned char *)&requestor->addr.s_addr)[3],
              requestor->port));
#endif

    _free_(requestor);
    return 0;
  }

  /*
   * SECURITY:
   * check if the friendship request has been previously announced.
   * if found, the pubkey of the foreign as previously
   * announced is returned, and we may consider to accept the request.
   * otherwise we refuse the friendship request.
   */

#ifdef DEBUG
  S(fprintf(stderr,
          "[%lu]\tfriendship request (peer %u.%u.%u.%u:%u): "
          "check if expected...\n",
          reqst->th_id,
          ((unsigned char *)&requestor->addr.s_addr)[0],
          ((unsigned char *)&requestor->addr.s_addr)[1],
          ((unsigned char *)&requestor->addr.s_addr)[2],
          ((unsigned char *)&requestor->addr.s_addr)[3],
          requestor->port));
#endif

  requestor->pubkey = search_expd_reqst(requestor->addr,requestor->port,
                                        pred_succ,reqst->th_id);

  if (requestor->pubkey == NULL) {

#ifdef DEBUG_F
    S(fprintf(stderr,"[%lu]\tnot expected\n",reqst->th_id));
#endif

    /*
     * SECURITY II:
     * Check if the friendship request comes from a peer in our routing tables
     * (neighbour, or finger).  This happens when RESponses travel back ("data"
     * messages).
     */

#ifdef DEBUG
    S(fprintf(stderr,
              "[%lu]\tcheck if request comes from a neighbour/finger...\n",
              reqst->th_id));
#endif

    if ((f = is_neighbour_or_finger(requestor)) != NULL) {

#ifdef DEBUG_F
      S(fprintf(stderr,"[%lu]\tfriendship req from a neighbour/finger\n",
                reqst->th_id));
#endif

      _RSA_copy_(requestor->pubkey,f->pubkey,goto abort);

    } else {

#ifdef DEBUG_F
      S(fprintf(stderr,"[%lu]\tnot from a neighbour/finger.  REFUSED\n",
                reqst->th_id));
#endif

      goto abort;
    }

  }

  /*  SECURITY III:  challenge requestor about its supposed pubkey.  */

  rc = _put_challenge_(reqst->csock,reqst->buf,requestor->pubkey);

  if (rc < 0) {

#ifdef DEBUG_F
    fprintf(stderr,"[%lu]\tchallenge interrupted\n",reqst->th_id);
#endif

    /*
     *  Maybe the requestor is just spoofing, trying to force this peer to
     *  cancel an expected friendship request.  Put the expected req back in
     *  queue, wait for a possible attempt from a legitimate peer.
     */
    if (requestor->pubkey == NULL)
      insert_expd_reqst(requestor->addr,requestor->port,pred_succ,
                        requestor->pubkey,reqst->th_id);

    goto abort;
  }

#ifdef DEBUG_F
  fprintf(stderr,"[%lu]\tchallenge ok\n",reqst->th_id);
#endif

  /*  OK, we can accept this new friend.  */

  _pthmutexinit_(&requestor->lock);
  requestor->in_cs = &reqst->csock;
  requestor->status = TOCALL;  /* become OK whenever opening f->cs outbound */
  if (pred_succ == 1) {
    /*
     *  This friend is our new predecessor in the ring.
     *  Flag it as such, remove such flag from old predecessor if any.
     *  Don't close now, because it might be in use by check_neighbours(),
     *  or needed for a RESponse to backtrack its corresponding REQuest.
     */
    requestor->from_predecessor = 1;
    forall_friends(f, if (f->from_predecessor) f->from_predecessor = 0;);
  }
  if (pred_succ == 2) {
    /*
     *  This friend is our new successor in the ring.
     *  Flag it as such, remove such flag from old successor if any.
     */
    requestor->from_successor = 1;
    forall_friends(f, if (f->from_successor) f->from_successor = 0;);
  }

  append_peer_list(&ring->friend_list,&ring->friend_list_tail,requestor);

  _write_(reqst->csock,"ok",strlen("ok")+1,rc,goto abort1);

#ifdef DEBUG
  S(fprintf(stderr,"[%lu]\t<< ok\n",reqst->th_id));
#endif

#ifdef DEBUG_F
  S(
  fprintf(stderr,"[%lu]\tfriendship req accepted",reqst->th_id);
  if (pred_succ == 1)
    fprintf(stderr,", from predecessor");
  if (pred_succ == 2)
    fprintf(stderr,", from successor");
  fprintf(stderr,"\n");
  fprintf(stderr,"[%lu]\tnew friend list:\n",reqst->th_id);
  forall_friends(f,
    fprintf(stderr,"[%lu]\t%u.%u.%u.%u:%u (%u) %s\n",
            reqst->th_id,
            ((unsigned char *)&f->addr.s_addr)[0],
            ((unsigned char *)&f->addr.s_addr)[1],
            ((unsigned char *)&f->addr.s_addr)[2],
            ((unsigned char *)&f->addr.s_addr)[3],
            f->port,f->status,
            f->from_predecessor ? "P" : f->from_successor ? "S" : " ");
  );
  );
#endif

  return 0;

abort:

  _free_(requestor);

abort1:

  _close_(&reqst->csock);
  return -1;

}
