/*********************************************************************/
/* queues.c                                                          */
/* routines for various queues/lists management                      */
/* see also queues.h                                                 */
/* see COPYING in parent directory                                   */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <queues.h>




/*
 * expected incoming friendship requests.
 */

expd_reqst_t      *expd_reqsts_head;    /* head of list of expected rqsts */
expd_reqst_t      *expd_reqsts_tail;    /* tail of list of expected rqsts */
unsigned int       expd_reqsts_howmany; /* how many elements in list      */
pthread_mutex_t    e_reqsts_lock;       /* mutex on expected reqsts list  */


void
init_expd_reqst_q
( void )
{

  /* the list of expected requests is empty initially */
  expd_reqsts_head = expd_reqsts_tail = NULL;
  expd_reqsts_howmany = 0;
  _pthmutexinit_(&e_reqsts_lock);
  return;

}



RSA *
search_expd_reqst
(
  struct in_addr addr, u_int16_t port, unsigned int pred_succ, pthread_t th_id
)
{

  expd_reqst_t         *pr;
  expd_reqst_t         *pr_prev;
  void                 *rc;
  unsigned int          dummy;


  dummy = th_id;
  pthread_mutex_lock(&e_reqsts_lock);

  pr_prev = pr = expd_reqsts_head;

#ifdef DEBUG_Q
  S(fprintf(stderr,"[%lu]\tsearch req: %u.%u.%u.%u:%u\n",
          th_id,
          ((unsigned char *)&addr.s_addr)[0],
          ((unsigned char *)&addr.s_addr)[1],
          ((unsigned char *)&addr.s_addr)[2],
          ((unsigned char *)&addr.s_addr)[3],
          port));
#endif

  /*  Search through the list by host addr, port, and "pred_succ" flag  */

  while (pr != NULL) {

#ifdef DEBUG_Q
    S(fprintf(stderr,"[%lu]\tpending rqst: %u.%u.%u.%u:%u\n",
              th_id,
              ((unsigned char *)&pr->foreign_addr.s_addr)[0],
              ((unsigned char *)&pr->foreign_addr.s_addr)[1],
              ((unsigned char *)&pr->foreign_addr.s_addr)[2],
              ((unsigned char *)&pr->foreign_addr.s_addr)[3],
              pr->foreign_port));
#endif

    /*  Equivalence between this host's IP addr and "localhost"  */
    if ((pr->foreign_addr.s_addr == addr.s_addr ||
        (pr->foreign_addr.s_addr == my_ipaddr_alt.s_addr &&
        addr.s_addr == my_ipaddr.s_addr) ||
        (pr->foreign_addr.s_addr == my_ipaddr.s_addr &&
        addr.s_addr == my_ipaddr_alt.s_addr))
        && (pr->foreign_port == port) && (pr->pred_succ == pred_succ))
      break;

    pr_prev = pr;
    pr = pr->next;

  }

  if (pr == NULL) {
    /* request from foreign node not found. no martini, no party! */

#ifdef DEBUG
    S(fprintf(stderr,"[%lu]\tsearch_expd_reqst(): NOT FOUND\n",th_id));
#endif

    pthread_mutex_unlock(&e_reqsts_lock);
    return NULL;
  }

  /* pending request from foreign node has been found */

#ifdef DEBUG
  S(
    fprintf(stderr,"[%lu]\tFOUND: %u.%u.%u.%u:%u",
            th_id,
            ((unsigned char *)&pr->foreign_addr.s_addr)[0],
            ((unsigned char *)&pr->foreign_addr.s_addr)[1],
            ((unsigned char *)&pr->foreign_addr.s_addr)[2],
            ((unsigned char *)&pr->foreign_addr.s_addr)[3],
            pr->foreign_port);
    if (pr->pred_succ == 1)
      fprintf(stderr,", from predecessor");
    if (pr->pred_succ == 2)
      fprintf(stderr,", from successor");
    fprintf(stderr,"\n");
  );
#endif

  /* delete the pending request from the list and dispose it */

#ifdef DEBUG_Q
  S(fprintf(stderr,"search_expd_reqst(): before removal: %u elements:\n",
          expd_reqsts_howmany));
  S(fprintf(stderr,"head %d, tail %d\n",
          (int)expd_reqsts_head,(int)expd_reqsts_tail));
  {expd_reqst_t *pr;
  for (pr = expd_reqsts_head; pr; pr = pr->next)
    S(fprintf(stderr,"%d ",(int)pr->timestamp.tv_sec));
  }
  S(fprintf(stderr,"\n"));
#endif

  expd_reqsts_howmany--;
  if (!expd_reqsts_howmany)
    expd_reqsts_head = expd_reqsts_tail = NULL;
  else
    if (pr == expd_reqsts_head)
      expd_reqsts_head = pr->next;
    else {
      if (pr == expd_reqsts_tail)
        expd_reqsts_tail = pr_prev;
      pr_prev->next = pr->next;
    }

#ifdef DEBUG_Q
  S(fprintf(stderr,"search_expd_reqst(): now %u elements:\n",
            expd_reqsts_howmany));
  S(fprintf(stderr,"head %d, tail %d\n",
            (int)expd_reqsts_head,(int)expd_reqsts_tail));
  {
    expd_reqst_t *pr;
    for (pr = expd_reqsts_head; pr; pr = pr->next)
      S(fprintf(stderr,"%d ",(int)pr->timestamp.tv_sec));
  }
  S(fprintf(stderr,"\n"));
#endif

  pthread_mutex_unlock(&e_reqsts_lock);

  rc = pr->pubkey;
  _free_(pr);
  return rc;

}



int
insert_expd_reqst
(
  struct in_addr requestor_addr, u_int16_t requestor_port,
  unsigned int pred_succ, RSA *pubkey, pthread_t th_id
)
{

  expd_reqst_t         *pr;
  unsigned int          dummy;
  

  dummy = th_id;
  _do_(_calloc_(pr,sizeof(expd_reqst_t)),!pr,"insert_expd_reqst(): pr calloc");
  pr->foreign_addr.s_addr = requestor_addr.s_addr;
  pr->foreign_port = requestor_port;
  pr->pred_succ = pred_succ;  /*  1 --> from pred;  2 --> from succ; 0 else  */
  pr->next = NULL;
  _RSA_copy_(pr->pubkey,pubkey,_free_(pr);return -1);
  gettimeofday(&(pr->timestamp),NULL);

  pthread_mutex_lock(&e_reqsts_lock);

#ifdef DEBUG_Q
  S(fprintf(stderr,"insert_expd_reqst(): %u elements:\n",expd_reqsts_howmany));
  S(fprintf(stderr,"head %d, tail %d\n",
          (int)expd_reqsts_head,(int)expd_reqsts_tail));
  S(fprintf(stderr,"timestamps: "));
  {expd_reqst_t *pr;
  for (pr = expd_reqsts_head; pr; pr = pr->next)
    S(fprintf(stderr,"%d ",(int)pr->timestamp.tv_sec));
  }
  S(fprintf(stderr,"\n"));
#endif

  expd_reqsts_howmany++;

#ifdef DEBUG_Q
  S(fprintf(stderr,
            "[%lu]\tinserting pending rqst, timestamp %d in list (now %u)\n",
            th_id,(int)pr->timestamp.tv_sec,expd_reqsts_howmany));
#endif

  if (expd_reqsts_tail != NULL) {
    expd_reqsts_tail->next = pr;
    expd_reqsts_tail = pr;
  }
  else
    expd_reqsts_head = expd_reqsts_tail = pr;

#ifdef DEBUG_Q
  S(fprintf(stderr,"insert_expd_reqst(): now %u elements:\n",
          expd_reqsts_howmany));
  S(fprintf(stderr,"head %d, tail %d\n",
          (int)expd_reqsts_head,(int)expd_reqsts_tail));
  S(fprintf(stderr,"timestamps: "));
  {expd_reqst_t *pr;
  for (pr = expd_reqsts_head; pr; pr = pr->next)
    S(fprintf(stderr,"%d ",(int)pr->timestamp.tv_sec));
  }
  S(fprintf(stderr,"\n"));
#endif

  pthread_mutex_unlock(&e_reqsts_lock);

  return 0;

}



void
remove_expd_reqst_from_successor
( pthread_t th_id )
{

  expd_reqst_t         *pr;
  expd_reqst_t         *pr_prev;
  unsigned int          dummy;


  dummy = th_id;
  pthread_mutex_lock(&e_reqsts_lock);

  pr_prev = pr = expd_reqsts_head;

  /*  Search through the list by pred_succ flag  */

  while (pr != NULL) {

#ifdef DEBUG_Q
    S(fprintf(stderr,"[%lu]\tpending rqst: %u.%u.%u.%u:%u\n",
              th_id,
              ((unsigned char *)&pr->foreign_addr.s_addr)[0],
              ((unsigned char *)&pr->foreign_addr.s_addr)[1],
              ((unsigned char *)&pr->foreign_addr.s_addr)[2],
              ((unsigned char *)&pr->foreign_addr.s_addr)[3],
              pr->foreign_port));
#endif

    if (pr->pred_succ == 2)
      break;

    pr_prev = pr;
    pr = pr->next;

  }

  if (pr == NULL) {
    /* request from successor not found. nothing to expunge */
    pthread_mutex_unlock(&e_reqsts_lock);
    return;
  }

  /* pending request from successor has been found.  delete and dispose. */

#ifdef DEBUG_Q
  S(
  fprintf(stderr,
          "remove_expd_reqst_from_successor(): before removal: %u elements:\n",
          expd_reqsts_howmany);
  fprintf(stderr,"head %d, tail %d\n",
          (int)expd_reqsts_head,(int)expd_reqsts_tail);
  {expd_reqst_t *pr;
  for (pr = expd_reqsts_head; pr; pr = pr->next)
    fprintf(stderr,"%d ",(int)pr->timestamp.tv_sec);
  }
  fprintf(stderr,"\n");
  );
#endif

  expd_reqsts_howmany--;
  if (!expd_reqsts_howmany)
    expd_reqsts_head = expd_reqsts_tail = NULL;
  else
    if (pr == expd_reqsts_head)
      expd_reqsts_head = pr->next;
    else {
      if (pr == expd_reqsts_tail)
        expd_reqsts_tail = pr_prev;
      pr_prev->next = pr->next;
    }

#ifdef DEBUG_Q
  S(
  fprintf(stderr,"remove_expd_reqst_from_successor(): now %u elements:\n",
          expd_reqsts_howmany);
  fprintf(stderr,"head %d, tail %d\n",
          (int)expd_reqsts_head,(int)expd_reqsts_tail);
  {
    expd_reqst_t *pr;
    for (pr = expd_reqsts_head; pr; pr = pr->next)
      fprintf(stderr,"%d ",(int)pr->timestamp.tv_sec);
  }
  fprintf(stderr,"\n");
  );
#endif

  pthread_mutex_unlock(&e_reqsts_lock);

  _RSA_free_(pr->pubkey);
  _free_(pr);
  return;

}



/*
 *  Garbage collector for the queue of expected requests.
 *  Runs periodically, as an autonomous thread.
 */
void *
clean_expd_reqsts
( void *data )
{

  expd_reqst_t          *pr,*pr_prev,*tpr;
  struct timeval         time_now;
  int                   *dummy;


  dummy = (int *)data;

  for (;;) {

    sleep(EXPD_PERIOD);

    gettimeofday(&time_now,NULL);

    pthread_mutex_lock(&e_reqsts_lock);

#ifdef DEBUG_Q
    S(fprintf(stderr,"clean_expd_reqsts(): %u elements:\n",
              expd_reqsts_howmany));
    S(fprintf(stderr,"head %d, tail %d\n",
              (int)expd_reqsts_head,(int)expd_reqsts_tail));
    S(fprintf(stderr,"timestamps: "));
    for (pr = expd_reqsts_head; pr; pr = pr->next)
      S(fprintf(stderr,"%d (%d)  ",
                (int)pr->timestamp.tv_sec,
                (int)time_now.tv_sec - (int)pr->timestamp.tv_sec));
    S(fprintf(stderr,"\n"));
#endif

    pr_prev = pr = expd_reqsts_head;

    while (pr != NULL) {
      if (((int)time_now.tv_sec - (int)pr->timestamp.tv_sec)
          <= (int)EXPD_DEADLINE)
        break;
      if (pr == expd_reqsts_head) {
        expd_reqsts_head = pr_prev = pr->next;
      } else {
        pr_prev->next = pr->next;
        if (pr == expd_reqsts_tail)
          expd_reqsts_tail = pr_prev;
      }
      expd_reqsts_howmany--;
      tpr = pr;
      pr = pr->next;
      _RSA_free_(tpr->pubkey);
      _free_(tpr);
    }
    if (expd_reqsts_head == NULL)
      expd_reqsts_tail = NULL;

#ifdef DEBUG_Q
    S(fprintf(stderr,"clean_expd_reqsts(): after clean: %u elements:\n",
              expd_reqsts_howmany));
    S(fprintf(stderr,"head %d, tail %d\n",
              (int)expd_reqsts_head,(int)expd_reqsts_tail));
    S(fprintf(stderr,"timestamps: "));
    for (pr = expd_reqsts_head; pr; pr = pr->next)
      S(fprintf(stderr,"%d (%d)  ",
                (int)pr->timestamp.tv_sec,
                (int)time_now.tv_sec - (int)pr->timestamp.tv_sec));
    S(fprintf(stderr,"\n"));
#endif

    pthread_mutex_unlock(&e_reqsts_lock);

  }  /*  Infinite loop  */

  return NULL;

}




/*
 * queue of pending read messages: issued but not yet answered.
 * max allowed queue size is MAX_P_READ
 */

pend_read_t       *pend_read;           /* queue, to be allocated at boot */
unsigned int       pend_read_head;
unsigned int       pend_read_tail;
unsigned int       pend_read_max;
pthread_mutex_t    p_read_tail_lock;    /* protect both "pend_read_tail"
                                           and "pend_read_max"            */


void
init_pend_read_q
( void )
{

  /* alloc the (initially empty) queue of pending read messages */

  /* this is an initial max.  will increase if needed */
  pend_read_max = I_MAX_P_READ;

  _do_(_calloc_(pend_read,pend_read_max*sizeof(pend_read_t)),!pend_read,
       "neblo_boot(): pend_read calloc");
  /*
   * pend_read[i].in_use and pend_read[i].done are automatically set to 0
   * by the calloc().  All handles are 0 as well.
   */

  pend_read_head = pend_read_tail = 0;

  _pthmutexinit_(&p_read_tail_lock);

  return;

}



int
insert_pend_read
(
  neblo_port port, neblo_addr addr, void *data, unsigned int maxlen,
  peer_t *from, neblo_rhandle from_rhandle, neblo_rhandle *rhandle
)
{

  unsigned int   i;
  pend_read_t   *pr;


  if (*rhandle != R_NONE) {
    pr = &pend_read[*rhandle];
    if (pr->in_use) {
      /* pending read msg already exists.  increment ref. count */
      /* where to put the i-th response when it arrives... */
      pr->data[pr->in_use] = data;
      pr->in_use++;
      return 0;
    }
  }

  pthread_mutex_lock(&p_read_tail_lock);

search:

  i = pend_read_tail;
  pend_read_tail = (pend_read_tail+1) % pend_read_max;

  if (pend_read_tail != pend_read_head) {
    /* found room in the queue.  fill it up... */
    pr = &pend_read[i];

    /* sanity check */
    if (pr->in_use) fprintf(stderr,"insert_pend_read(): BUG!!\n");

    pr->handle = (neblo_rhandle)i;
    pr->from = from;
    pr->from_rhandle = from_rhandle;
    pr->port = port;
    pr->addr = addr;
    /* where to put the 1-st response when it arrives... */
    if (pr->in_use < NEBLO_MAXCP)
      pr->data[pr->in_use] = data;
    pr->in_use++;
    pr->maxlen = maxlen;
    _pthmutexinit_(&pr->lock);
    _pthmutexinit_(&pr->wait_sem);
    _pthcondinit_(&pr->wait_cond);
    gettimeofday(&(pr->timestamp),NULL);

    pthread_mutex_unlock(&p_read_tail_lock);
    *rhandle = (neblo_rhandle)i;
    return 0;
  }

  /* queue is full.  "enlarge your penis" then search a free entry again... */
  pend_read_tail = i;
  if (pend_read_max >= MAX_P_READ) {
    /* could not enlarge.  insert fails... */
    pthread_mutex_unlock(&p_read_tail_lock);
    return -1;
  }
  pend_read_max <<= 1;
  _realloc_(pr,pend_read,pend_read_max*sizeof(pend_read_t));
  if (!pr) {
    /* could not enlarge.  insert fails... */
    pend_read_max >>= 1;
    pthread_mutex_unlock(&p_read_tail_lock);
    return -1;
  }
  pend_read = pr;
  goto search;

}



pend_read_t *
search_pend_read
( neblo_rhandle h )
{

  pend_read_t   *pr;


  if (h != R_NONE) {
    pr = &pend_read[h];
    if (pr->in_use && pr->handle == h)
      /* pending read found */
      return pr;
  }
  return NULL;

}



void
mark_pend_read_done
( neblo_rhandle h )
{

  pend_read_t   *pr;


  if (h != R_NONE) {
    pr = &pend_read[h];
    pthread_mutex_lock(&pr->lock);  
    if (pr->in_use && pr->handle == h)
      pr->done++;
    pthread_mutex_unlock(&pr->lock);  
  }
  return;

}



void
cancel_pend_read
( pend_read_t *pr )
{

  if (pr != NULL) {
    pr->in_use = 0;
    pr->done = 0;
  }
  return;

}



/*
 *  Garbage collector for the queue of pending read messages.
 *  Runs periodically, as an autonomous thread.
 *  Queue of pending reads is a vector not list, so will not free memory.
 */
void *
clean_pend_reads
( void *data )
{

  struct timeval        time_now;
  pend_read_t          *pr;
  int                  *dummy;


  dummy = (int *)data;

  for (;;) {

    sleep(PEND_READ_PERIOD);

    gettimeofday(&time_now,NULL);

    for (; pend_read_head != pend_read_tail;
           pend_read_head = (pend_read_head+1) % pend_read_max) {
      pr = &pend_read[pend_read_head];
      pthread_mutex_lock(&pr->lock);
      if (!pr->in_use) {
        pthread_mutex_unlock(&pr->lock);
        continue;
      }
      if (((int)time_now.tv_sec - (int)pr->timestamp.tv_sec)
          <= (int)PEND_READ_DEADLINE) {
        pthread_mutex_unlock(&pr->lock);
        break;
      }
      pr->in_use = 0;
      pr->done = 0;
      if (pr->from == NULL)
        /* wake up local waiting peer, which will realize read has timed out */
        pthread_cond_signal(&pr->wait_cond);
      pthread_mutex_unlock(&pr->lock);
    }

  }  /*  Infinite loop  */

  return NULL;

}



/*
 * cancel pending reads coming from a given peer.
 * necessary when a peer dies and is deleted from routing tables
 */
void
clean_pend_reads_peer
( peer_t *f )
{

  unsigned int          i;
  pend_read_t          *pr;


  for (i = pend_read_head; i != pend_read_tail; i = (i+1) % pend_read_max) {
    pr = &pend_read[i];
    if (pend_read[i].from == f) {
      pthread_mutex_lock(&pr->lock);
      pend_read[i].in_use = 0;
      pend_read[i].done = 0;
      pthread_mutex_unlock(&pr->lock);
    }
  }

  return;

}




/*
 * queue of read/write/data msgs that could not be forwarded yet, and thus
 * must be retried
 */

retry_reqst_t     *retry_queue_head;    /* head of list of expected rqsts */
retry_reqst_t     *retry_queue_tail;    /* tail of list of expected rqsts */
unsigned int       retry_queue_howmany; /* how many elements in list      */
pthread_mutex_t    retry_queue_lock;    /* mutex on expected reqsts list  */


void
init_retry_queue
( void )
{

  /* the list of is empty initially */
  retry_queue_head = retry_queue_tail = NULL;
  retry_queue_howmany = 0;
  _pthmutexinit_(&retry_queue_lock);
  return;

}



int
insert_retry_queue
( rwd_t kind, reqst_t *reqst, pthread_t th_id )
{

  retry_reqst_t        *pr;
  unsigned int          dummy;
  

  dummy = th_id;

  _do_(_calloc_(pr,sizeof(retry_reqst_t)),!pr,
       "insert_retry_queue(): pr calloc");
  pr->kind = kind;
  pr->reqst = reqst;
  pr->pending = 1;
  pr->next = NULL;
  gettimeofday(&(pr->timestamp),NULL);

  pthread_mutex_lock(&retry_queue_lock);

#ifdef DEBUG_Q
  S(fprintf(stderr,"insert_retry_queue(): %u elements:\n",
            retry_queue_howmany));
  S(fprintf(stderr,"head %d, tail %d\n",
          (int)retry_queue_head,(int)retry_queue_tail));
  S(fprintf(stderr,"timestamps: "));
  {retry_reqst_t *pr;
  for (pr = retry_queue_head; pr; pr = pr->next)
    S(fprintf(stderr,"%d ",(int)pr->timestamp.tv_sec));
  }
  S(fprintf(stderr,"\n"));
#endif

  retry_queue_howmany++;

#ifdef DEBUG_Q
  S(fprintf(stderr,
          "[%lu]\tinserting rwd pend rqst of kind %u, timestamp %d in list (now %u)\n",
          th_id,kind,(int)pr->timestamp.tv_sec,retry_queue_howmany));
#endif

  if (retry_queue_tail != NULL) {
    retry_queue_tail->next = pr;
    retry_queue_tail = pr;
  }
  else
    retry_queue_head = retry_queue_tail = pr;

#ifdef DEBUG_Q
  S(fprintf(stderr,"insert_retry_queue(): now %u elements:\n",
          retry_queue_howmany));
  S(fprintf(stderr,"head %d, tail %d\n",
          (int)retry_queue_head,(int)retry_queue_tail));
  S(fprintf(stderr,"timestamps: "));
  {retry_reqst_t *pr;
  for (pr = retry_queue_head; pr; pr = pr->next)
    S(fprintf(stderr,"%d ",(int)pr->timestamp.tv_sec));
  }
  S(fprintf(stderr,"\n"));
#endif

  pthread_mutex_unlock(&retry_queue_lock);

  return 0;

}



/*
 *  Garbage collector for the queue of delayed msgs.
 *  Runs periodically, as an autonomous thread.
 */
void *
clean_retry_queue
( void *data )
{

  retry_reqst_t        *pr;
  retry_reqst_t        *pr_prev;
  retry_reqst_t        *tpr;
  int                   rc = 0;
  struct timeval        time_now;
  int                  *dummy;


  dummy = (int *)data;

  for (;;) {

    sleep(RETRY_PERIOD);

    gettimeofday(&time_now,NULL);

    pthread_mutex_lock(&retry_queue_lock);

#ifdef DEBUG_Q
    S(fprintf(stderr,"clean_retry_queue(): %u elements:\n",
              retry_queue_howmany));
    S(fprintf(stderr,"head %d, tail %d\n",
             (int)retry_queue_head,(int)retry_queue_tail));
    S(fprintf(stderr,"timestamps: "));
    for (pr = retry_queue_head; pr; pr = pr->next)
      S(fprintf(stderr,"%d (%d)  ",
              (int)pr->timestamp.tv_sec,
              (int)time_now.tv_sec - (int)pr->timestamp.tv_sec));
    S(fprintf(stderr,"\n"));
#endif

    pr_prev = pr = retry_queue_head;

    while (pr != NULL) {

      if (((int)time_now.tv_sec - (int)pr->timestamp.tv_sec)
          > (int)RETRY_DEADLINE) {

        /*
         *  This delayed msg is expired.
         *  Give last chance, but delete from queue.
         */

        if (pr == retry_queue_head) {
          retry_queue_head = pr_prev = pr->next;
        } else {
          pr_prev->next = pr->next;
          if (pr == retry_queue_tail)
            retry_queue_tail = pr_prev;
        }
        retry_queue_howmany--;
        tpr = pr;
        pr = pr->next;

        /*
         *  We now try to serve the delayed msg "tpr".  This may require
         *  sending additional read/write messages, which in turn might
         *  be inserted in the retry queue in case of routing failure.
         *  We thus must release the retry lock to avoid a possible deadlock.
         */
        pthread_mutex_unlock(&retry_queue_lock);

        if (tpr->pending) {
          /* this retry reqst was not managed in due time. last chance! */

          pthread_mutex_lock(&clean_lock);
          dont_clean++;
          pthread_mutex_unlock(&clean_lock);

          switch(tpr->kind) {
            case WRITE:
              neblo_handle_UD_msgs(tpr->reqst,LAST_CHANCE);
              break;
            case READ:
              neblo_handle_REQ_msgs(tpr->reqst,LAST_CHANCE);
              break;
            case DATA:
              neblo_handle_RES_msgs(tpr->reqst);
              break;
            case REFR:
              neblo_handle_refresh_msgs(tpr->reqst,LAST_CHANCE);
              break;
          }

          pthread_mutex_lock(&clean_lock);
          dont_clean--;
          pthread_mutex_unlock(&clean_lock);

        }

        _free_(tpr->reqst->buf);
        _free_(tpr->reqst);
        _free_(tpr);

        /* back to the retry queue management */
        pthread_mutex_lock(&retry_queue_lock);
        continue;
  
      }

      /*
       *  Retry serving this delayed msg.
       *  Release the retry lock to avoid a possible deadlock.
       *  See comment above.
       */

      pthread_mutex_unlock(&retry_queue_lock);

      pthread_mutex_lock(&clean_lock);
      dont_clean++;
      pthread_mutex_unlock(&clean_lock);

      switch(pr->kind) {
        case WRITE:
          rc = neblo_handle_UD_msgs(pr->reqst,RETRY);
          break;
        case READ:
          rc = neblo_handle_REQ_msgs(pr->reqst,RETRY);
          break;
        case DATA:
          rc = neblo_handle_RES_msgs(pr->reqst);
          break;
        case REFR:
          rc = neblo_handle_refresh_msgs(pr->reqst,RETRY);
          break;
      }

      pthread_mutex_lock(&clean_lock);
      dont_clean--;
      pthread_mutex_unlock(&clean_lock);

      /* back to the retry queue management */
      pthread_mutex_lock(&retry_queue_lock);

      if (rc >= 0) {
        /* operation was successful this time, so delete from queue */
        if (pr == retry_queue_head) {
          retry_queue_head = pr_prev = pr->next;
        } else {
          pr_prev->next = pr->next;
          if (pr == retry_queue_tail)
            retry_queue_tail = pr_prev;
        }
        retry_queue_howmany--;
        tpr = pr;
        pr = pr->next;
        _free_(tpr->reqst->buf);
        _free_(tpr->reqst);
        _free_(tpr);
      } else {
        /*
         * operation still unsuccessful but not yet timed out,
         * so leave in queue
         */
        pr_prev = pr;
        pr = pr->next;
      }

    } /* end while */

    if (retry_queue_head == NULL)
      retry_queue_tail = NULL;

#ifdef DEBUG_Q
    S(fprintf(stderr,"clean_retry_queue(): after clean: %u elements:\n",
              retry_queue_howmany));
    S(fprintf(stderr,"head %d, tail %d\n",
              (int)retry_queue_head,(int)retry_queue_tail));
    S(fprintf(stderr,"timestamps: "));
    for (pr = retry_queue_head; pr; pr = pr->next)
      S(fprintf(stderr,"%d (%d)  ",
                (int)pr->timestamp.tv_sec,
                (int)time_now.tv_sec - (int)pr->timestamp.tv_sec));
    S(fprintf(stderr,"\n"));
#endif

    pthread_mutex_unlock(&retry_queue_lock);

  }  /*  Infinite loop  */

  return NULL;

}
